﻿using System;
using System.Collections.Generic;
using Celeste;
using Celeste.Mod;
using Microsoft.Xna.Framework;

namespace TileLayoutAnchor;

public class TileLayoutAnchorMapDataProcessor : EverestMapDataProcessor{
	public struct TileSubstitution{
		public string Room;
		// position *within* the room
		public Vector2 AnchorOffset;

		public bool ControlTiles, ControlTileEntities, ControlEffects, ControlDecals;

		public override string ToString(){
			return $"({Room}, {AnchorOffset})";
		}
	}

	public static readonly Dictionary<string /* SID */, List<TileSubstitution>> Substitutions = new();

	private string roomName = null;

	public override void Reset() => Substitutions.Remove(AreaKey.GetSID());

	public override Dictionary<string, Action<BinaryPacker.Element>> Init(){
		return new(){
			{
				"level", level => {
					// be sure to write the level name down.
					roomName = level.Attr("name").Split(':')[0];
					if(roomName.StartsWith("lvl_"))
						roomName = roomName[4..];
				}
			},
			{
				"entity:TileLayoutAnchor/TileLayoutAnchor", anchor => {
					TileLayoutAnchorModule.LogDebug($"Found a layout anchor in room {roomName}, SID {AreaKey.GetSID()}!");
					// if we haven't encountered this level yet, give it an empty entry
					Substitutions.TryAdd(AreaKey.GetSID(), []);
					Substitutions[AreaKey.GetSID()].Add(new TileSubstitution{
						Room = roomName,
						// note that this is position *within* a room
						AnchorOffset = new Vector2(anchor.AttrFloat("x"), anchor.AttrFloat("y")),
						ControlTiles = anchor.AttrBool("controlTiles", true),
						ControlTileEntities = anchor.AttrBool("controlTileEntities", true),
						ControlEffects = anchor.AttrBool("controlEffects", true),
						ControlDecals = anchor.AttrBool("controlDecals", true)
					});
				}
			}
		};
	}

	public override void End(){}
}